-- cparser: parsing C/C++ cleaned head files 
-- Waldemar Celes
-- Mar 1997


$debug

-- Global table to store typedefs
TYPEDEF = {}

-- Global table to store user tags
USERTAG = {n=0}


-- BASIC PATTERN MATCHING

-- end of class
ENDCLASS = format('%c', 1)

-- ArgumentList: ( ... )
--> return the arguments declaration without the paranthesis.
p_argumentlist = "%s*%(([^%)]*)%)"

-- Valid operators: []
p_operator = "%[? ?%]?+?-?%*?/?<?>?=?"

-- PATTERN TABLE

parseTable = {
   -- Directive: $...
   --> pattern return the line
    {"directive", "^%s*$([^\n]*\n)"},

   -- Enum: enum { ... };
   --> pattern returns the enumeration body.
    {"enum", "^%s*enum%s*{([^}]*)}"..ENDCLASS.."?%s*;"},
   
   -- Define: #define name [value] \n
   --> pattern returns the constant name.
    {"define", "^%s*#define%s%s*([^%s]*)[^\n]*[^\n]*\n"},

   -- Function: Fullname ArgumentList 
   --> pattern returns the function full name, the argument list, 
   --> and if const (applied for methods only)
    {"func", "^%s*([%w%_~<>][%w%*&:_%s<>]*"..p_operator..")%s*"
             ..p_argumentlist.."%s*(c?o?n?s?t?)%s*;[^\n]*\n"
    },
   
   -- Class: class Name [inheritance] { ... };
   --    or: struct Name [inheritance] { ... };
   --> pattern returns the class name, the inheritace body, and the class body.
    {"class", "^%s*class%s*([^%s:{]*)([^{]*){([^"..
      ENDCLASS.."]*)}"..ENDCLASS.."%s*;"},
    {"class", "^%s*struct%s*([^%s:{]*)([^{]*){([^"..
      ENDCLASS.."]*)}"..ENDCLASS.."%s*;"},

   -- Typedef (for variables): typedef Name ...;
   --> pattern returns the fullname (type+name)
    {"typedef", "^%s*typedef%s*([%w%*&_][%w%*&_%s]*)%s*;[^\n]*\n"},

   -- Typedef (for functions): typedef Function,
   -- where the function name is given by: (*name)
   --> pattern returns the fulltype, the name, and the argument list. 
    {"funcdef", "^typedef%s*([%w_][%w%*&_%s]*)%(%s*%*%s*([%w_][%w_]*)%s*%)%s*"..
             p_argumentlist..";[^\n]*\n"},
   
   -- Variable: fullname;
   --> pattern returns the full name (fulltype+name)
    {"variable", "^%s*([%w%*&_][%w%*&_%s]*)%s*;[^\n]*\n"},

   -- Blank line(s)
    {"blank", "^%s%s*"}
}

-- Function "What is?"
-- The following values are returned:
-- what is matched, a table listing the substring returned by the matching,
-- and the remaining string.
-- Returns nil if the string doesn't match with anything. 
-- Note: it's assumed number of returned values per pattern does not exceed 5.
function whatIs (string)
 local i=1
 while parseTable[i] do
  local b,e,f1,f2,f3,f4,f5 = strfind(string,parseTable[i][2])
  if b and e then
   return parseTable[i][1],{f1,f2,f3,f4,f5},strsub(string,e+1)
  end
  i = i+1
 end
 if strfind(string,"[^%s]") then
  -- restore ENDCLASS mark
  string = gsub(string,ENDCLASS,"")
  error("toLua parse error after "..toLua_debug_message.."\n\n"..
        "Current buffered code is:\n"..
        strsub(string,1,100)
       )
 end
 return nil
end


-- PARSE FUNCTIONS

-- Given a string and a delimiter, returns a list of tokens (a table) 
-- present in the string.
function extractTokens (string,delimiter)
 -- Substitute delimiter by my own delimiter
 local s = gsub(string,delimiter,"$")
 tokenTable = {n=0}
 gsub(s,"%s*$?%s*([^$][^$]*)%s*",storeToken)
 return tokenTable
end 

-- Function to support extractToken operation
function storeToken (token)
 tokenTable.n = tokenTable.n+1
 tokenTable[tokenTable.n] = token
end

-- Parse fulltype
-- Fulltype is a list of modifier followed by a type.
-- The function returns three strings: modifiers, type, and pointer.
function parseFullType (fulltype)
 local tokens = extractTokens(fulltype,"%s%s*")
 local type = tokens[tokens.n]
 local modifiers = ""
 local i=1
 while i ~= tokens.n do
  if modifiers == "" then modifiers=tokens[i]
  else modifiers = modifiers .." ".. tokens[i] end
  i = i+1
 end
 modifiers = gsub(modifiers,"virtual","")
 modifiers = gsub(modifiers,"%s%s*"," ")
 modifiers = gsub(gsub(modifiers,"^%s%s*",""),"%s%s*$","")
 type = gsub(gsub(type,"^%s%s*",""),"%s%s*$","")
 if type ~= "void*" and type ~= "char*" and 
    strsub(type,strlen(type)) == "*" 
 then
  pointer = "*"
  type = strsub(type,1,strlen(type)-1)
 elseif  strsub(type,strlen(type)) == "&" then
  pointer = "&"
  type = strsub(type,1,strlen(type)-1)
 else
  pointer = ""
 end

 -- check if type is a typedef
 while TYPEDEF[type] do
  if TYPEDEF[type].modifiers ~= "" then
   modifiers = modifiers .." "..TYPEDEF[type].modifiers
  end
  pointer = pointer..TYPEDEF[type].pointer
  type = TYPEDEF[type].type
 end

 if type == "" and pointer ~= "" then
  modifiers,type,pointer = parseFullType(modifiers..pointer) 
  if type == "" then
   error ('toLua parse error: "'..fulltype..' ".\n') 
  end
 end

 if strfind(modifiers,"operator") then
  error('toLua parse error: "'..fulltype.. 
     ' ".\n     Do not use space when specifying operator.')
 end

 return modifiers, type, pointer
end

-- Parse fullname
-- Fullname is a fulltype (possibly blank) followed by a name.
-- The function returns three strings: modifiers, type, and name.
function parseFullName (fullname, isVariable)
 local fulltype, name, pointer = parseFullType(fullname) -- they are all names

 name = name..pointer  -- to deal with 'operator*'

 if strsub(name,1,1) == "*" then 
  name = strsub(name,2)
  fulltype = fulltype.."*"
 elseif strsub(name,1,1) == "&" then
  name = strsub(name,2)
  fulltype = fulltype.."&"
 end
 local modifiers, type = "", ""
 if fulltype ~= "" then
  modifiers, type, pointer = parseFullType(fulltype)
 end
 
 local default = ""
 local dimension = ""
 local domain = ""

 if isVariable then
  -- extract default value, if any (used for parameter names)
  local tokens = extractTokens(name,"=") 
  name = tokens[1]
  if tokens[2] then 
   default = tokens[2] 
  end
 
  -- extract dimension value, if it is a pre-defined vector (used for parameters)
  local b,e,n,d = strfind(name,"([^%[]*)%[(.*)%]")
  if b and e and n and d then
   name = n
   dimension = d
  end
 
  -- convert pointer to basic type into an array of dimension 1
  if type ~= "" and TYPE[type] and pointer ~= "" then
   if dimension ~= "" then
    error("toLua - array of pointer to basic types is not allowed.")
   end
   pointer = ""
   dimension = "1"
  end

 else
  -- extract name domain, if any (used for method names)
  local tokens = extractTokens(name,"::")
  if tokens[2] then
   domain = tokens[1]
   name = tokens[2]
  else
   name = tokens[1]
  end

  -- cannot return a pointer to basic type
  if TYPE[type] and pointer ~= "" then
   -- except reference in operator[]
   if not (name == "operator[]" and pointer == "&") then  
    error ("toLua - function cannot return a pointer to a basic type.")
   end
  end

 end
  
 return modifiers, type, pointer, domain, name, default, dimension
end

-- Parse argument list
-- The function returns a table containing the fullnames of the arguments.
-- Each item in the table is composed by: 
-- modifiers, type, name,default,dimension.
-- If the list is void, the first and unique _type_ in the table is "void".
function parseArgumentList (argList)
 local args = extractTokens(argList,",")
 if not args[1] or (strfind(args[1],"^%s*void%s*$") and args.n == 1) then
  return {{modifiers="",type="void",pointer="",name="",default="",dimension=""}}
 end
 local i=1
 local argTable = {}
 while i <= args.n do
  local modifiers,type,pointer,domain,name,default,dimension = 
        parseFullName(args[i],1)
  argTable[i] = {modifiers=modifiers, type=type, pointer=pointer, 
                 name=name, default=default,dimension=dimension
                }
  i = i+1
 end
 return argTable
end 

-- Parse a function
-- As input, it expects a table with the function fullname and 
-- the argument list. 
-- function when that results in a function.
-- Returns a table with:
--    modifiers = function type modifier;
--    type = function type;
--    name = function name;
--    args = table as returned by function "parseArgumentList"
function parseFunction (fields)
 local modifiers,type,pointer,domain,name = parseFullName(fields[1],nil)
 local args = parseArgumentList(fields[2])
 toLua_debug_message = "function: "..name
 return {modifiers=modifiers, type=type, pointer=pointer, 
         domain=domain, name=name, args=args, const=fields[3]
        }
end

-- Parse a define
-- As input, it expects a table with the constant name.
-- Returns the constant name received.
function parseDefine (fields)
 toLua_debug_message = "define: "..fields[1]
 return fields[1]
end

-- Parse a enum
-- As input, it expects a table with the enum body.
-- Returns a list of constant defined in the enum.
function parseEnum (fields)
 local tokens = extractTokens(fields[1],",")
 local i=1
 local constants = {}
 while tokens[i] do
  local b,e,const = strfind(tokens[i],"^%s*([%w_][%w_]*)") 
  if b and e then
   constants[i] = const
  end
  i = i+1
 end
 toLua_debug_message = "enum: "..constants[1].."..."
 return constants
end

-- Parse a variable
-- As input, it expects a table with the variable full name.
-- Returns a table the modifiers, the type, and the variable name.
function parseVariable (fields)
 local modifiers, type, pointer, domain, name = parseFullName(fields[1],1)
 return {modifiers=modifiers, type=type, pointer=pointer, name=name}
end

-- Parse a typedef
-- As input, it expects a table with the type fullname.
-- Returns a table the modifiers, the type, and the type name.
function parseTypedef (fields)
 local t = parseVariable(fields)
 -- add to global typedef table
 TYPEDEF[t.name] = t
 toLua_debug_message = "typedef: "..t.name
 return
end

-- Parse a funcdef
-- As input, it expects a table with the fulltype, the name, and the
-- argument list.
-- Returns a table with the modifiers, the type, the name, and the
-- argument table as returned by "parseArgumentList"
function parseFuncdef (fields)
 local modifiers,type, pointer = parseFullType(fields[1])
 local name = fields[2]
 local args = parseArgumentList(fields[3])
 return {modifiers=modifiers, type=type, pointer=pointer, name=name, args=args }
end

-- Parse class
-- As input, it expects a table with the class name, the inheritance body,
-- and the class body.
-- Returns a table with:
--    name = class name;
--    base = base class name;
--    body = table as returned by function "parseChunk"
function parseClass (fields)
 local name = fields[1]
 local b,e,base = strfind(fields[2],"%s*:%s*public%s%s*([%w_][%w_]*)")
 if not base then base = "" end
 local body = parseChunk(fields[3])
 toLua_debug_message = "class: "..name
 return {name=name,base=base,body=body}
end

-- Parse a chunk
-- As input, it expects the chunk body or a table with the chunk body as 
-- its first element.
-- Returns a table that list all constants (defines and enums), variables, 
-- and functions of the chunk.
-- The has has the following information:
-- chunk = {
--      constants = {NAME1,NAME2,...},
--      variables = {{modifiers=..., type=..., pointer=..., name=...},
--                   ...
--                  },
--      functions = {{modifiers=..., type=..., pointer=..., name=name,
--                    args={{modifiers=..., type=..., pointer=..., name=...},
--                          {modifiers=..., type=..., pointer=..., name=...},
--                           ...
--                          }
--                         }
--                   },
--                   ...
--                  }
--      classes = {{name=...,base=...,body = {...}},  -- body is a chunk
--                 ...
--                }
--     }
function parseChunk (body)
 if type(body) ~= "string" then body = body[1] end

 local chunk = {typedefs={n=0},
                funcdefs={n=0},
                constants={n=0},
                variables={n=0},
                functions={n=0},
                classes={n=0}
               }
 local item, fields, body = whatIs(body)
 while item do
  if     item == "directive" then
   write(fields[1])	-- write the line as it is
  elseif item == "typedef" then
   chunk.typedefs.n = chunk.typedefs.n + 1
   chunk.typedefs[chunk.typedefs.n] = parseTypedef(fields)
  elseif item == "funcdef" then
   chunk.funcdefs.n = chunk.funcdefs.n + 1
   chunk.funcdefs[chunk.funcdefs.n] = parseFuncdef(fields)
  elseif item == "variable" then
   chunk.variables.n = chunk.variables.n + 1
   chunk.variables[chunk.variables.n] = parseVariable(fields)
  elseif item == "define" then
   chunk.constants.n = chunk.constants.n + 1
   chunk.constants[chunk.constants.n] = parseDefine(fields)
  elseif item == "enum" then
   local i=1
   local list = parseEnum(fields)
   while list[i] do
    chunk.constants.n = chunk.constants.n + 1
    chunk.constants[chunk.constants.n] = list[i]
    i = i+1
   end
  elseif item == "func" then
   chunk.functions.n = chunk.functions.n + 1
   chunk.functions[chunk.functions.n] = parseFunction(fields)
  elseif item == "class" then
   chunk.classes.n = chunk.classes.n + 1
   chunk.classes[chunk.classes.n] = parseClass(fields)
  elseif item == "blank" then
   -- do nothing
  elseif body then
   error(body)
  end
  item, fields, body = whatIs(body)
 end
 return chunk
end

-- PRINT CHUNK'S FIELDS
-- For debug purpose only

function printChunk (title, t)
 local i

 -- output title
 print(title)

 -- output typedefs
 if t.typedefs.n > 0 then
  print("\n\nTYPEDEFS")
  i=1
  while t.typedefs[i] do
   print("$"..t.typedefs[i].modifiers..
         "$"..t.typedefs[i].type..
         "$"..t.typedefs[i].name..
         "$"
   )
   i = i+1
  end
 end
 
 -- output funcdefs
 if t.funcdefs.n > 0 then
  print("\n\nFUNCDEFS")
  i=1
  while t.funcdefs[i] do
   print("$"..t.funcdefs[i].modifiers..
         "$"..t.funcdefs[i].type..
         "$"..t.funcdefs[i].name..
         "$"
   )
   print("Parameters:")
   local j=1
   while t.funcdefs[i].args[j] do
    print("$"..t.funcdefs[i].args[j].modifiers..
          "$"..t.funcdefs[i].args[j].type..
          "$"..t.funcdefs[i].args[j].name..
          "$"..t.funcdefs[i].args[j].default..
         "$"
    )
    j = j+1
   end
   print("")
   i = i+1
  end
 end
  
 -- output constants
 if t.constants.n > 0 then
  print("\n\nCONSTANTS")
  i=1
  while t.constants[i] do
   print(t.constants[i])
   i = i+1
  end
 end
 
 -- output variables
 if t.variables.n > 0 then
  print("\n\nVARIABLES")
  i=1
  while t.variables[i] do
   print("$"..t.variables[i].modifiers..
         "$"..t.variables[i].type..
         "$"..t.variables[i].name..
         "$"
   )
   i = i+1
  end
 end
 
 -- output functions
 if t.functions.n > 0 then
  print("\n\nFUNCTIONS")
  i=1
  while t.functions[i] do
   print("$"..t.functions[i].modifiers..
         "$"..t.functions[i].type..
         "$"..t.functions[i].domain..
         "$"..t.functions[i].name..
         "$"
   )
   print("Parameters:")
   local j=1
   while t.functions[i].args[j] do
    print("$"..t.functions[i].args[j].modifiers..
          "$"..t.functions[i].args[j].type..
          "$"..t.functions[i].args[j].name..
          "$"..t.functions[i].args[j].default..
         "$"
    )
    j = j+1
   end
   print("")
   i = i+1
  end
 end

  -- output classes
 if t.classes.n > 0 then
  print("\n\nCLASSES")
  i=1
  while t.classes[i] do
   printChunk("$"..t.classes[i].name.."$"..t.classes[i].base.."$",
              t.classes[i].body
             )
   i = i+1
  end
 end
end 

