/*
** Support for automatic lua interface code generation 
** Waldemar Celes
** Aug 1997
*/

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <assert.h>

#ifdef __cplusplus
extern "C" {
#endif
#include "lua.h"
#ifdef __cplusplus
}
#endif

#include "tolua.h"

/* Reserve tag numbers to represent basic types */
int TOLUA_NUMBER_TAG;
int TOLUA_STRING_TAG;
int TOLUA_USERDATA_TAG;

typedef void (*Getfb) (void *);
typedef void (*Setfb) (void *, lua_Object);
typedef int (*GetIndexfb) (lua_Object, lua_Object);
typedef int (*SetIndexfb) (lua_Object, lua_Object, lua_Object);


/* Internal function */
static void setInternalMethods (int tag);


/*
** Error messages
*/
static char errmsg[100];
void toLua_wrongself (char* funcname)
{
 sprintf(errmsg,"toLua - 'self' is not appropriate to function '%s'.",
         funcname
        );
 lua_error(errmsg);
}
void toLua_wrongarg (char* funcname, int narg, char* tname)
{
 sprintf(errmsg,"toLua - argument %d of function '%s' should be of type '%s'.",
         narg,funcname,tname
        );
 lua_error(errmsg);
}
void toLua_wrongarray (char* funcname, int narg, char* tname, int dim)
{
 sprintf(errmsg,"toLua - argument %d of function '%s'\n     should be a %d-dimension array of '%s'.",
         narg,funcname,dim,tname
        );
 lua_error(errmsg);
}
void toLua_toomanyargs (char* funcname)
{
 sprintf(errmsg,"toLua - too many arguments to function '%s'.",
         funcname
        );
 lua_error(errmsg);
}


/*
** Create global tables
*/

static int tagRef;       /* typeTag  = TABLE[typeName] */
static int nameRef;      /* typeName = TABLE[typeTag]  */
static int parentRef;    /* parentTag = TABLE[baseTag] */
static int castRef;      /* castFunc = TABLE[tag] */
static int allocRef;     /* TABLE[pointer] = 1 */
static int getRef;       /* getFallback = TABLE[tag] */
static int setRef;       /* setFallback = TABLE[tag] */
static int mateRef;      /* mateTable = TABLE[pointer] */
static int addRef;       /* string: "operator_add" */
static int subRef;       /* string: "operator_sub" */
static int mulRef;       /* string: "operator_mul" */
static int divRef;       /* string: "operator_div" */
static int ltRef;        /* string: "operator_lt" */
static int gtRef;        /* string: "operator_gt" */
static int leRef;        /* string: "operator_le" */
static int geRef;        /* string: "operator_ge" */
static int getindexRef;  /* string: "operator_index_get" */
static int setindexRef;  /* string: "operator_index_set" */

static void createReferences (void)
{
 lua_Object lt;
 lua_beginblock();

 lt = lua_createtable();
 lua_pushobject(lt);
 lua_setglobal("toLua_tag");
 lua_pushobject(lt);
 tagRef = lua_ref(1);

 lt = lua_createtable();
 lua_pushobject(lt);
 lua_setglobal("toLua_name");
 lua_pushobject(lt);
 nameRef = lua_ref(1);

 lt = lua_createtable();
 lua_pushobject(lt);
 lua_setglobal("toLua_parent");
 lua_pushobject(lt);
 parentRef = lua_ref(1);

 lt = lua_createtable();
 lua_pushobject(lt);
 castRef = lua_ref(1);

 lt = lua_createtable();
 lua_pushobject(lt);
 allocRef = lua_ref(1);

 lt = lua_createtable();
 lua_pushobject(lt);
 setRef = lua_ref(1);

 lt = lua_createtable();
 lua_pushobject(lt);
 getRef = lua_ref(1);

#if 1
 lt = lua_createtable();
 lua_pushobject(lt);
 lua_setglobal("toLua_mate");
 lua_pushobject(lt);
 mateRef = lua_ref(1);
#else
 lt = lua_createtable();
 lua_pushobject(lt);
 mateRef = lua_ref(1);
#endif

 lua_pushstring("operator_add");
 addRef = lua_ref(1);
 lua_pushstring("operator_sub");
 subRef = lua_ref(1);
 lua_pushstring("operator_mul");
 mulRef = lua_ref(1);
 lua_pushstring("operator_div");
 divRef = lua_ref(1);
 lua_pushstring("operator_lt");
 ltRef = lua_ref(1);
 lua_pushstring("operator_gt");
 gtRef = lua_ref(1);
 lua_pushstring("operator_le");
 leRef = lua_ref(1);
 lua_pushstring("operator_ge");
 geRef = lua_ref(1);
 lua_pushstring("operator_index_get");
 getindexRef = lua_ref(1);
 lua_pushstring("operator_index_set");
 setindexRef = lua_ref(1);

 lua_endblock();
}

/*
** Reserve basic tags
*/
void basicTags (void)
{
 /* just ask lua some uniques identification values */
 TOLUA_NUMBER_TAG = lua_newtag();
 TOLUA_STRING_TAG = lua_newtag();
 TOLUA_USERDATA_TAG = lua_newtag();
}

/*
** Provide a new, unused tag, given a type name.
*/
int toLua_newtag(char *type)
{
 int t=0;
 toLua_init();	/* because a tag may be required before initialization */
 if (type != NULL)
 {
  lua_Object tag;
  lua_beginblock(); 
  lua_pushref(tagRef);
  lua_pushstring(type);
  tag = lua_gettable();
  if (lua_isnumber(tag))
  {
   t = (int)lua_getnumber(tag);
  }
  else
  {
   t = lua_newtag();
   lua_pushref(tagRef);
   lua_pushstring(type);
   lua_pushnumber(t);
   lua_settable();
   lua_pushref(nameRef);
   lua_pushnumber(t);
   lua_pushstring(type);
   lua_settable();
   setInternalMethods(t);
  }
  lua_endblock();
 }
 return t;
}

/*
** Register a parent
*/
void toLua_registerparent (int derived, int base, void* (*fcast)(void*))
{
 lua_beginblock();
 lua_pushref(parentRef);
 lua_pushnumber(derived);
 lua_pushnumber(base);
 lua_settable();
 lua_pushref(castRef);
 lua_pushnumber(derived);
 lua_pushuserdata((void*)fcast);
 lua_settable();
 lua_endblock();
}

/*
** Check if object is of a corresponding type
*/
int toLua_istype (lua_Object lo, int type)
{
 if (lo == LUA_NOOBJECT)
  return 0;
 if ((type==TOLUA_STRING_TAG || type==TOLUA_USERDATA_TAG) && lua_isnil(lo))
  return 1;
 else if (type == TOLUA_STRING_TAG)
  return lua_isstring(lo);
 else if (type == TOLUA_NUMBER_TAG)
  return lua_isnumber(lo);
 else if (type == TOLUA_USERDATA_TAG)
  return lua_isuserdata(lo);
 else
 {
  int objtag;
  if (lua_isnil(lo))
   return 1;
  lua_beginblock();
  objtag = lua_type(lo);
  while (objtag != type && objtag != 0)    /* assume 0 as the default tag */
  {
   /* access parent */
   lua_Object pobj;
   lua_pushref(parentRef);
   lua_pushnumber(objtag);
   pobj = lua_gettable();
   if (!lua_isnumber(pobj))
   {
    lua_endblock();
    return 0;
   }
   objtag = (int)lua_getnumber(pobj);
  }
  lua_endblock();
  return 1;
 } 
}

int toLua_nistype (int index, int type)
{
 lua_Object lo = lua_getparam(index);
 return toLua_istype(lo,type);
}

int toLua_nisarray (int index, int type, int dim)
{
 int i;
 int res;
 lua_Object lo = lua_getparam(index);
 if (!lua_istable(lo))
  return 0;
 lua_beginblock();
 for (i=1; i<=dim; i++)
 {
  if (!toLua_istype(lua_getindexed(lo,i),type))
  {
   lua_endblock();
   return 0;
  }
 }
 res = lua_isnil(lua_getindexed(lo,i)) ? 1 : 0;
 lua_endblock();
 return res;
}

/*
** Functions to get Lua values
*/
lua_Object toLua_gettable (lua_Object o)
{
 return o;
}
char *toLua_getstring (lua_Object o)
{
 if (lua_isnil(o))
  return NULL;
 else
  return lua_getstring(o);
}
float toLua_getnumber (lua_Object o)
{
 return lua_getnumber(o);
}
lua_Object toLua_getfunction (lua_Object o)
{
 return o;
}
void *toLua_getuserdata (lua_Object o)
{
 if (lua_isnil(o))
  return NULL;
 else
  return lua_getuserdata(o);
}
void *toLua_getusertag (lua_Object o, int tag)
{
 void* res;
 int objtag;
 if (lua_isnil(o))
  return NULL;
 lua_beginblock();
 res = lua_getuserdata(o);
 objtag = lua_type(o);
 while (objtag != tag && objtag != 0)	/* assume 0 as the default tag */
 {
  /* access parent */
  lua_Object pobj, cfunc;
  lua_pushref(parentRef);
  lua_pushnumber(objtag);
  pobj = lua_gettable();
  lua_pushref(castRef);
  lua_pushnumber(objtag);
  cfunc = lua_gettable();
  assert(lua_isnumber(pobj)&&lua_isuserdata(cfunc));
  res = ((void* (*)(void*))lua_getuserdata(cfunc))(res); /* convert to base */
  objtag = (int)lua_getnumber(pobj);
 }
 lua_endblock();
 return res;
}

/*
** Internal method to collect object passed by value to Lua
*/
static void toLua_fbgc (void)
{
 lua_Object lo;
 void* v = lua_getuserdata(lua_getparam(1));
 lua_pushref(allocRef);
 lua_pushuserdata(v);
 lo = lua_rawgettable();
 if (!lua_isnil(lo)) 
 {
  /* eliminate this entry from the allocation table */
  lua_pushref(allocRef);
  lua_pushuserdata(v);
  lua_pushnil();
  lua_rawsettable();
  free(v);
 }
}

/*
** Function to push values
*/
void toLua_pushstring (char* s)
{
 if (s == NULL)
  lua_pushnil();
 else
  lua_pushstring(s); 
}

void toLua_pushusertag (void *v, int tag)
{
#if 0
 if (v == NULL)
  lua_pushnil();
 else
  lua_pushusertag(v,tag);
#else
 if (v == NULL)
  lua_pushnil();
 else
 {
  lua_Object lo;
  lua_pushusertag(v,LUA_ANYTAG);
  lo = lua_pop();
  if (lua_tag(lo) == 0)
  {
   lua_pushobject(lo);
   lua_settag(tag);
  }
  lua_pushobject(lo); 
 }
#endif
}

/*
** Dynamically allocate a copy of a given object and register
** the allocated address so that the garbage collection
** function can deallocate it.
*/
void toLua_pushuservalue (void* v, int size, int tag)
{
 void *m = malloc(size);

 lua_beginblock();
 lua_pushref(allocRef);
 lua_pushuserdata(m);
 lua_pushnumber(1);
 lua_settable();
 lua_pushcfunction(toLua_fbgc);
 lua_settagmethod(tag,"gc"); /* set (or reset) tag method */
 lua_endblock();

 memcpy(m,v,size);
 lua_pushusertag(m,tag);
}

/***********************************************************************
** Function to register internal methods
*/

/*
** Return the Lua table used to store fallbacks of a given type (tag).
*/
static lua_Object toLua_gettagtable (int ref, int tag)
{
 lua_Object lo;
 lua_pushref(ref);
 lua_pushnumber(tag);
 lo = lua_gettable();
 if (!lua_istable(lo))
 {
  lo = lua_createtable();
  lua_pushref(ref);
  lua_pushnumber(tag);
  lua_pushobject(lo);
  lua_settable();
 }
 return lo;
} 

/*
** Register a gettable fallback.
*/
void toLua_getfb (int tag, char *f, void *fb)
{
 lua_Object lo;
 lua_beginblock();
 lo = toLua_gettagtable(getRef,tag); 
 lua_pushobject(lo);
 lua_pushstring(f);
 lua_pushuserdata(fb);
 lua_settable();
 lua_endblock();
}

/*
** Register a settable fallback.
*/
void toLua_setfb (int tag, char *f, void *fb)
{
 lua_Object lo;
 lua_beginblock();
 lo = toLua_gettagtable(setRef,tag);
 lua_pushobject(lo);
 lua_pushstring(f);
 lua_pushuserdata(fb);
 lua_settable();
 lua_endblock();
}

/*
** Register a function fallback.
*/
void toLua_funcfb (int tag, char *f, lua_CFunction cfunc)
{
 lua_Object lo;
 lua_beginblock();
 lo = toLua_gettagtable(getRef,tag);
 lua_pushobject(lo);
 lua_pushstring(f);
 lua_pushcfunction(cfunc);
 lua_settable();
 lua_endblock();
}

/*
** Set fallbacks to access external variables: uses the "lib" name as a table
*/
void toLua_libtag (char *lib)
{
 int tag;
 lua_beginblock();
 tag = toLua_newtag(lib);
 lua_pushusertag(NULL,tag);
 lua_setglobal(lib);
 lua_endblock();
}
void toLua_getfb_extern (char *lib, char *var, void *fb)
{
 int tag = toLua_newtag(lib);
 toLua_getfb (tag, var, fb);
}
void toLua_setfb_extern (char *lib, char *var, void *fb)
{
 int tag = toLua_newtag(lib);
 toLua_setfb (tag, var, fb);
}

/*
** Return the registered fallback (function to be called). 
*/
static lua_Object toLua_getregistredfb (int ref, lua_Object t, lua_Object i)
{
 lua_Object lo;
 int objtag = lua_tag(t);
 lua_pushobject(toLua_gettagtable(ref,objtag));
 lua_pushobject(i);
 lo = lua_gettable();
 while (lua_isnil(lo))
 {
  /* access base */
  lua_Object pobj;
  lua_pushref(parentRef);
  lua_pushnumber(objtag);
  pobj = lua_gettable();
  if (!lua_isnumber(pobj))
   break;                  /* no (more) base */
  objtag = (int)lua_getnumber(pobj);
  lua_pushobject(toLua_gettagtable(ref,objtag));
  lua_pushobject(i);
  lo = lua_gettable();
 }
 return lo;
}

/*
** Return Lua table associated with a C object.
*/
static lua_Object toLua_getmatetable (lua_Object cobj)
{
 lua_pushref(mateRef);
 lua_pushobject(cobj);
 return lua_gettable();
}

/*
** Create mate table
*/
static lua_Object toLua_creatematetable (lua_Object cobj)
{
 lua_Object lo = lua_createtable();
 lua_pushref(mateRef);
 lua_pushobject(cobj);
 lua_pushobject(lo);
 lua_settable();
 return lo;
}

/*
** Tag methods to overload operators
*/
static void toLua_op_add (void)
{
 lua_Object lo1 = lua_getparam(1); 
 lua_Object lo2 = lua_getparam(2); 
 lua_Object lo;
 lo = toLua_getregistredfb(getRef,lo1,lua_getref(addRef));
 if (!lua_iscfunction(lo))
  lua_error("toLua - undefined 'add' operator");
 lua_pushobject(lo1);
 lua_pushobject(lo2);
 lua_callfunction(lo);
 lua_pushobject(lua_getresult(1));
}
static void toLua_op_sub (void)
{
 lua_Object lo1 = lua_getparam(1);
 lua_Object lo2 = lua_getparam(2);
 lua_Object lo;
 lo = toLua_getregistredfb(getRef,lo1,lua_getref(subRef));
 if (!lua_iscfunction(lo))
  lua_error("toLua - undefined 'sub' operator");
 lua_pushobject(lo1);
 lua_pushobject(lo2);
 lua_callfunction(lo);
 lua_pushobject(lua_getresult(1));
}
static void toLua_op_mul (void)
{
 lua_Object lo1 = lua_getparam(1);
 lua_Object lo2 = lua_getparam(2);
 lua_Object lo;
 lo = toLua_getregistredfb(getRef,lo1,lua_getref(mulRef));
 if (!lua_iscfunction(lo))
  lua_error("toLua - undefined 'mul' operator");
 lua_pushobject(lo1);
 lua_pushobject(lo2);
 lua_callfunction(lo);
 lua_pushobject(lua_getresult(1));
}
static void toLua_op_div (void)
{
 lua_Object lo1 = lua_getparam(1);
 lua_Object lo2 = lua_getparam(2);
 lua_Object lo;
 lo = toLua_getregistredfb(getRef,lo1,lua_getref(divRef));
 if (!lua_iscfunction(lo))
  lua_error("toLua - undefined 'div' operator");
 lua_pushobject(lo1);
 lua_pushobject(lo2);
 lua_callfunction(lo);
 lua_pushobject(lua_getresult(1));
}
static void toLua_op_lt (void)
{
 lua_Object lo1 = lua_getparam(1);
 lua_Object lo2 = lua_getparam(2);
 lua_Object lo;
 lo = toLua_getregistredfb(getRef,lo1,lua_getref(ltRef));
 if (!lua_iscfunction(lo))
  lua_error("toLua - undefined 'lt' operator");
 lua_pushobject(lo1);
 lua_pushobject(lo2);
 lua_callfunction(lo);
 lo = lua_getresult(1);
 if (lua_isnumber(lo) && lua_getnumber(lo)==0)
  lua_pushnil();
 else
  lua_pushobject(lo);
}
static void toLua_op_gt (void)
{
 lua_Object lo1 = lua_getparam(1);
 lua_Object lo2 = lua_getparam(2);
 lua_Object lo;
 lo = toLua_getregistredfb(getRef,lo1,lua_getref(gtRef));
 if (!lua_iscfunction(lo))
  lua_error("toLua - undefined 'gt' operator");
 lua_pushobject(lo1);
 lua_pushobject(lo2);
 lua_callfunction(lo);
 lo = lua_getresult(1);
 if (lua_isnumber(lo) && lua_getnumber(lo)==0)
  lua_pushnil();
 else
  lua_pushobject(lo);
}
static void toLua_op_le (void)
{
 lua_Object lo1 = lua_getparam(1);
 lua_Object lo2 = lua_getparam(2);
 lua_Object lo;
 lo = toLua_getregistredfb(getRef,lo1,lua_getref(leRef));
 if (!lua_iscfunction(lo))
  lua_error("toLua - undefined 'le' operator");
 lua_pushobject(lo1);
 lua_pushobject(lo2);
 lua_callfunction(lo);
 lo = lua_getresult(1);
 if (lua_isnumber(lo) && lua_getnumber(lo)==0)
  lua_pushnil();
 else
  lua_pushobject(lo);
}
static void toLua_op_ge (void)
{
 lua_Object lo1 = lua_getparam(1);
 lua_Object lo2 = lua_getparam(2);
 lua_Object lo;
 lo = toLua_getregistredfb(getRef,lo1,lua_getref(geRef));
 if (!lua_iscfunction(lo))
  lua_error("toLua - undefined 'ge' operator");
 lua_pushobject(lo1);
 lua_pushobject(lo2);
 lua_callfunction(lo);
 lo = lua_getresult(1);
 if (lua_isnumber(lo) && lua_getnumber(lo)==0)
  lua_pushnil();
 else
  lua_pushobject(lo);
}

static int toLua_op_getindex (lua_Object t, lua_Object i)
{
 lua_Object lo = toLua_getregistredfb(getRef,t,lua_getref(getindexRef));
 if (!lua_isuserdata(lo))
  return 0;
 else
 {
  GetIndexfb fb = (GetIndexfb)lua_getuserdata(lo);
  return fb(t,i);
 }
}

static int toLua_op_setindex (lua_Object t, lua_Object i, lua_Object v)
{
 lua_Object lo = toLua_getregistredfb(setRef,t,lua_getref(setindexRef));
 if (!lua_isuserdata(lo))
  return 0;
 else
 {
  SetIndexfb fb = (SetIndexfb)lua_getuserdata(lo);
  return fb(t,i,v);
 }
}

/*
** "gettable" fallback.
*/
static void toLua_fbgettable (void)
{
 lua_Object t = lua_getparam(1);
 lua_Object i = lua_getparam(2);
 lua_Object table = toLua_getmatetable(t);
 if (lua_istable(table))
 {
  lua_Object obj;
  lua_pushobject(table);
  lua_pushobject(i);
  obj = lua_gettable();
  if (!lua_isnil(obj))
  {
   lua_pushobject(obj);
   return;
  }
 }
 /* no mate table or field not present in mate table */
 if (!toLua_op_getindex(t,i))
 {
  lua_Object lo;
  lua_beginblock();
  lo = toLua_getregistredfb(getRef,t,i);
  if (lua_isuserdata(lo))
  {
   void* fb = lua_getuserdata(lo);
   lua_endblock();
   ((Getfb)fb)(lua_getuserdata(t));
  }
  else if (lua_iscfunction(lo))
  {
   lua_CFunction fb = lua_getcfunction(lo);
   lua_endblock();
   lua_pushcfunction(fb); /* return method */
  }
  else
  {
   lua_endblock();
   lua_pushnil(); /* always consider as a it was a table */
  }
 }
}

/*
** "settable" fallback.
*/
static void toLua_fbsettable (void)
{
 lua_Object t = lua_getparam(1);
 lua_Object i = lua_getparam(2);
 lua_Object v = lua_getparam(3);
 lua_Object lo;
 lua_beginblock();
 lo = toLua_getregistredfb(setRef,t,i);
 if (lua_isuserdata(lo))
 {
  void* fb = lua_getuserdata(lo);
  ((Setfb)fb)(lua_getuserdata(t),v);
 }
 else if (!toLua_op_setindex(t,i,v))
 {
  lua_Object table = toLua_getmatetable(t);
  if (!lua_istable(table))
   table = toLua_creatematetable(t);
  lua_pushobject(table);
  lua_pushobject(i); 
  lua_pushobject(v); 
  lua_settable();
 } 
 lua_endblock();
}

/*
** Set internal methods
*/
static void setInternalMethods (int tag)
{
 lua_pushcfunction(toLua_fbgettable);
 lua_settagmethod(tag,"gettable");
 lua_pushcfunction(toLua_fbsettable);
 lua_settagmethod(tag,"settable");
 lua_pushcfunction(toLua_op_add);
 lua_settagmethod(tag,"add");
 lua_pushcfunction(toLua_op_sub);
 lua_settagmethod(tag,"sub");
 lua_pushcfunction(toLua_op_mul);
 lua_settagmethod(tag,"mul");
 lua_pushcfunction(toLua_op_div);
 lua_settagmethod(tag,"div");
 lua_pushcfunction(toLua_op_lt);
 lua_settagmethod(tag,"lt");
 lua_pushcfunction(toLua_op_gt);
 lua_settagmethod(tag,"gt");
 lua_pushcfunction(toLua_op_le);
 lua_settagmethod(tag,"le");
 lua_pushcfunction(toLua_op_ge);
 lua_settagmethod(tag,"ge");
}

/*
** Function to initialize fallbacks
*/
void toLua_init (void)
{
 static int first = 1;
 if (first)
 {
  createReferences();
  basicTags();
  first = 0;
 }
}
